let currentTabIndex = -1;
let isAltPressed = false;
const KEY_PRESS_THRESHOLD = 50;
const tabSound = new Audio(chrome.runtime.getURL('soundfile.mp3'));
const VIRTUAL_LIST_BUFFER = 10;
const lastKeyPressTime = 0;
const lastFlickTime = 0;
const FLICK_THRESHOLD = 100;

class VirtualList {
  constructor(container, itemHeight = 60) {
    this.container = container;
    this.itemHeight = itemHeight;
    this.items = [];
    this.visibleItems = new Map();
    this.scrollTop = 0;
    this.lastRenderTime = 0;
    this.previousActiveTabId = null;
    this.isInitialRender = true;
  }

  setItems(items) {
    if (this.activeTabId) {
      this.previousActiveTabId = this.activeTabId;
    }
    
    // If this is the first render, create all elements
    if (this.isInitialRender) {
      this.items = items;
      this.container.style.height = `${items.length * this.itemHeight}px`;
      this.render();
      this.isInitialRender = false;
    } else {
      // Otherwise, just update the data and reposition existing elements
      this.items = items;
      this.updatePositions();
    }
  }

  updatePositions() {
    const itemCount = this.items.length;
    
    this.items.forEach((item, index) => {
      const element = this.visibleItems.get(index);
      
      if (element) {
        const favicon = item.favIconUrl ? `<img src="${item.favIconUrl}" class="tab-favicon" />` : '';
        element.innerHTML = `
          <div class="tab-content">
            ${favicon}
            <span class="tab-title">${item.title}</span>
            ${item.id === this.activeTabId ? '' : '<span class="hover-indicator">Switch</span>'}
          </div>
        `;

        // Calculate position relative to the selected item (index 0)
        const relativeIndex = index;
        const yOffset = relativeIndex * 60;
        const zOffset = -Math.abs(relativeIndex) * 40;
        const rotation = relativeIndex * 8;
        const scale = Math.max(0.85, 1 - Math.abs(relativeIndex) * 0.075);
        
        // Update visual properties with transitions
        element.style.opacity = Math.max(0.4, 1 - Math.abs(relativeIndex) * 0.15);
        element.style.zIndex = 1000 - Math.abs(relativeIndex);
        
        // Apply transform with smooth transition
        element.style.transform = `
          translateX(-50%)
          translateY(${yOffset}px)
          translateZ(${zOffset}px)
          rotateX(${rotation}deg)
          scale(${scale})`;
          
        // Update selected state
        if (index === 0) {
          element.classList.add('selected');
        } else {
          element.classList.remove('selected');
        }
      }
    });
  }

  render() {
    if (performance.now() - this.lastRenderTime < 16) return;
    this.lastRenderTime = performance.now();

    const totalItems = this.items.length;
    const visibleIndexes = new Set();

    // Create elements for all items if they don't exist
    for (let i = 0; i < totalItems; i++) {
      visibleIndexes.add(i);

      if (!this.visibleItems.has(i)) {
        const item = this.items[i];
        const element = this.createItemElement(item, i);
        this.container.appendChild(element);
        this.visibleItems.set(i, element);
      }
    }

    // Update positions for all visible items
    this.updatePositions();
  }

  createItemElement(tab, index) {
    const tabItem = document.createElement('div');
    tabItem.className = 'tab-item';
    if (tab.id === this.activeTabId) {
      tabItem.classList.add('current-tab');
    }
    if (index === currentTabIndex) {
      tabItem.classList.add('selected');
    }

    const favicon = tab.favIconUrl ? `<img src="${tab.favIconUrl}" class="tab-favicon" />` : '';
    tabItem.innerHTML = `
      <div class="tab-content">
        ${favicon}
        <span class="tab-title">${tab.title}</span>
        ${tab.id === this.activeTabId ? '' : '<span class="hover-indicator">Switch</span>'}
      </div>
    `;

    tabItem.addEventListener('click', () => this.onItemClick(tab, index));
    return tabItem;
  }

  onItemClick(tab, index) {
    chrome.runtime.sendMessage({ action: 'switchTab', tabId: tab.id });
    const overlay = document.getElementById('tab-flicker-overlay');
    if (overlay) {
      overlay.classList.remove('visible');
      setTimeout(() => overlay.remove(), 200); // Slightly faster removal
    }
  }

  updateScroll(scrollTop) {
    this.scrollTop = scrollTop;
    requestAnimationFrame(() => this.render());
  }
}

let virtualList;
let originalTabs = [];  // Store the original tab order

document.addEventListener('keydown', (event) => {
  const overlay = document.getElementById('tab-flicker-overlay');

  if (event.altKey && (event.key.toLowerCase() === 'q' || event.key === 'œ')) {
    event.preventDefault();

    if (!isAltPressed) {
      isAltPressed = true;
      toggleOverlay();
    } else {
      cycleNextTab();
    }
  } else if (overlay && event.key.toLowerCase() === 'q') {
    event.preventDefault();
    cycleNextTab();
  }
});

document.addEventListener('keyup', (event) => {
  if (event.key === 'Alt') {
    isAltPressed = false;
    const selectedTab = virtualList?.items?.[currentTabIndex];
    if (selectedTab) {
      chrome.runtime.sendMessage({ action: 'switchTab', tabId: selectedTab.id });
      const overlay = document.getElementById('tab-flicker-overlay');
      if (overlay) {
        overlay.classList.remove('visible');
        setTimeout(() => overlay.remove(), 200);
      }
    }
    currentTabIndex = -1;
    resetTabHighlights();
  } else if (event.key.toLowerCase() === 'q' && !isAltPressed && virtualList?.items?.length > 0) {
    const selectedTab = virtualList?.items?.[currentTabIndex];
    if (selectedTab) {
      chrome.runtime.sendMessage({ action: 'switchTab', tabId: selectedTab.id });
      const overlay = document.getElementById('tab-flicker-overlay');
      if (overlay) {
        overlay.classList.remove('visible');
        setTimeout(() => overlay.remove(), 200);
      }
    }
  }
});

function resetTabHighlights() {
  for (const el of document.querySelectorAll('.tab-item')) {
    el.classList.remove('selected');
  }
}

function cycleNextTab() {
  const items = virtualList?.items || [];
  if (items.length === 0) return;
  
  // Get the current tab's ID
  const currentTab = items[0];
  // Find its index in the original tabs array
  const originalIndex = originalTabs.findIndex(tab => tab.id === currentTab.id);
  
  // Reorder tabs starting from the next tab in the original order
  const reorderedTabs = [
    // Next tab becomes the first (the one after current in original order)
    ...originalTabs.slice(originalIndex + 1),
    // Wrap around to the beginning up to current
    ...originalTabs.slice(0, originalIndex + 1)
  ];
  
  // Update the virtual list with the reordered array
  virtualList.setItems(reorderedTabs);
  currentTabIndex = 0;
  virtualList.activeTabId = reorderedTabs[0].id;
  
  // Remove previous selected states and add to current
  for (const el of document.querySelectorAll('.tab-item')) {
    el.classList.remove('selected');
  }
  const selectedElement = virtualList.visibleItems.get(0);
  if (selectedElement) {
    selectedElement.classList.add('selected');
  }
  
  // Play sound effect
  tabSound.currentTime = 0;
  tabSound.play();
}

function createOverlay() {
  const overlay = document.createElement('div');
  overlay.id = 'tab-flicker-overlay';
  overlay.innerHTML = `
    <div class="overlay-content">
      <div class="tabs-container"></div>
    </div>
  `;
  document.body.appendChild(overlay);

  const tabsContainer = overlay.querySelector('.tabs-container');
  return overlay;
}

document.addEventListener('keydown', (event) => {
  if (event.key === 'Escape') {
    const overlay = document.getElementById('tab-flicker-overlay');
    if (overlay) {
      overlay.classList.remove('visible');
      setTimeout(() => overlay.remove(), 200); // Slightly faster removal
    }
  }
});

function toggleOverlay() {
  let overlay = document.getElementById('tab-flicker-overlay');

  if (overlay) {
    overlay.classList.remove('visible');
    setTimeout(() => {
      overlay.remove();
      currentTabIndex = -1;
    }, 200);
  } else {
    overlay = createOverlay();
    overlay.classList.add('visible');
    
    const tabsContainer = overlay.querySelector('.tabs-container');
    virtualList = new VirtualList(tabsContainer);

    chrome.runtime.sendMessage({ action: 'getTabs' }, (response) => {
      if (response.tabs) {
        const tabs = response.tabs;
        originalTabs = [...tabs];
        const activeTabIndex = tabs.findIndex(tab => tab.active);
        
        // Reorder tabs: active tab first, followed by tabs to its right, then wrap around to the beginning
        const reorderedTabs = [
          tabs[activeTabIndex],
          ...tabs.slice(activeTabIndex + 1),
          ...tabs.slice(0, activeTabIndex)
        ];

        virtualList.setItems(reorderedTabs);
        currentTabIndex = 0;
        virtualList.activeTabId = reorderedTabs[0].id;
        
        tabSound.currentTime = 0;
        tabSound.play();
      } else {
        console.error("Failed to fetch tabs");
        overlay.remove();
      }
    });

    overlay.addEventListener('click', (event) => {
      if (event.target === overlay) {
        overlay.classList.remove('visible');
        setTimeout(() => overlay.remove(), 200);
      }
    });
  }
}

// Listen for tab activation
chrome.runtime.onMessage.addListener((request, sender, sendResponse) => {
  if (request.action === 'tabActivated') {
    // Reset the extension state
    isAltPressed = false;
    currentTabIndex = -1;
    virtualList = null;
    const overlay = document.getElementById('tab-flicker-overlay');
    if (overlay) {
      overlay.remove();
    }
  }
});

console.log('Tab Flicker content script loaded');